#include <limits.h>
#include <errno.h>
#include <ihklib.h>
#include "util.h"
#include "okng.h"
#include "params.h"
#include "linux.h"
#include <unistd.h>

const char param[] = "dev_index";
const char *values[] = {
	"INT_MIN",
	"-1",
	"0",
	"1",
	"INT_MAX",
};

int main(int argc, char **argv)
{
	int ret;
	int i;

	params_getopt(argc, argv);

	/* Precondition */
	ret = linux_insmod(0);
	INTERR(ret, "linux_insmod returned %d\n", ret);

	int dev_index_input[] = {
		INT_MIN,
		-1,
		0,
		1,
		INT_MAX
	};

	int ret_expected[5] = {
		-ENOENT,
		-ENOENT,
		0,
		-ENOENT,
		-ENOENT,
	};

	int ret_expected_os_instances[5] = {
		0,
		0,
		1,
		0,
		0,
	};

	for (i = 0; i < 5; i++) {
		int os_index;

		START("test-case: %s: %s\n", param, values[i]);

		ret = ihk_create_os(dev_index_input[i]);
		OKNG(ret == ret_expected[i],
		     "return value (os index when positive): %d, expected: %d\n",
		     ret, ret_expected[i]);
		os_index = ret;

		ret = ihk_get_num_os_instances(0);
		OKNG(ret == ret_expected_os_instances[i],
		     "# of os instances: %d, expected: %d\n",
		     ret, ret_expected_os_instances[i]);

		if (ihk_get_num_os_instances(0)) {
			ret = ihk_destroy_os(dev_index_input[i], os_index);
		}
	}

	ret = 0;
out:
	linux_rmmod(0);
	return ret;
}

